"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
Object.defineProperty(exports, "observe", {
  enumerable: true,
  get: function () {
    return _observerUtil.observe;
  }
});
Object.defineProperty(exports, "unobserve", {
  enumerable: true,
  get: function () {
    return _observerUtil.unobserve;
  }
});
exports.Component = exports.sharedComponentData = void 0;

var _react = require("react");

var _observerUtil = require("@nx-js/observer-util");

const testEnv = typeof process == 'object' && process != null && typeof process.env == 'object' && process.env != null && process.env.NODE_ENV == 'test';
const bindComponentExcludes = ['constructor', 'render', // 'componentDidMount', must be bound in case this function is called in a component method
'shouldComponentUpdate', 'getSnapshotBeforeUpdate', 'componentDidUpdate', // 'componentWillUnmount', must be bound in case this function is called in a component method
'componentDidCatch', // 'mounted', must be bound in case this function is called in a component method
'updated' // 'beforeUnmount' must be bound in case this function is called in a component method
];
const bindInstanceExcludes = ['constructor', '__defineGetter__', '__defineSetter__', 'hasOwnProperty', '__lookupGetter__', '__lookupSetter__', 'isPrototypeOf', 'propertyIsEnumerable', 'toString', 'valueOf', '__proto__', 'toLocaleString'];
const observableExcludes = ['props', 'context', 'refs', 'updater', 'state'];

const sharedComponentData = object => {
  let shared = (0, _observerUtil.observable)(object); // Bind methods

  for (const name of Object.getOwnPropertyNames(Object.getPrototypeOf(object))) {
    if (!bindInstanceExcludes.includes(name)) shared[name] = shared[name].bind(shared);
  }

  return shared;
};

exports.sharedComponentData = sharedComponentData;

// The same observable must be used elsewhere or else components will not rerender
class Component extends _react.Component {
  // Returns the last instance
  static instance() {
    if (!Component._instances) return null;

    const instances = Component._instances.get(this.name);

    if (!instances || instances.length == 0) return null;
    return instances[instances.length - 1];
  }

  static instances() {
    if (!Component._instances) return [];

    const instances = Component._instances.get(this.name);

    if (!instances) return [];
    return instances;
  }

  constructor(...args) {
    super(...args);
    if (!Component._instances) Component._instances = new Map();

    let instances = Component._instances.get(this.constructor.name);

    if (!instances) Component._instances.set(this.constructor.name, [this]);else instances.push(this); // Bind methods

    for (const name of Object.getOwnPropertyNames(Object.getPrototypeOf(this))) {
      if (!name.startsWith('UNSAFE_component') && !bindComponentExcludes.includes(name)) {
        this[name] = this[name].bind(this);
      }
    } // Create getters and setters for properties. Objects are made observable.


    const render = this.render;
    let variables = new Map();
    let rerenderVariables = new Set();

    if (testEnv) {
      this._reactSimplifiedVariables = variables;
      this._reactSimplifiedRerenderVariables = rerenderVariables;
    }

    let firstRenderCall = true;
    let duringRenderCall = false;

    this.render = () => {
      if (firstRenderCall) {
        firstRenderCall = false;
        const ownPropertyNames = Object.getOwnPropertyNames(this);

        for (const name of ownPropertyNames) {
          const type = typeof this[name];

          if (type != 'function' && !name.startsWith('_react') && !observableExcludes.includes(name)) {
            if (type == 'undefined') {
              throw new Error('Detected undefined member variable ' + name + ' in class ' + this.constructor.name + '. Since such variables are unsafe to use in render(), you must initialize this variable.');
            }

            if (type == 'object' && this[name] != null) {
              // Do not make React.createRef() objects observable
              if (Object.isSealed(this[name])) {
                let keys = Object.keys(this[name]);
                if (keys.length == 1 && keys[0] == 'current') continue;
              }

              variables.set(name, (0, _observerUtil.observable)(this[name]));
            } else {
              variables.set(name, this[name]);
            }

            Object.defineProperty(this, name, {
              get: () => {
                if (duringRenderCall) rerenderVariables.add(name);
                return variables.get(name);
              },
              set: value => {
                const rerender = rerenderVariables.has(name);
                variables.set(name, typeof value == 'object' && value != null && rerender ? (0, _observerUtil.observable)(value) : value);
                if (rerender) this.setState({});
              }
            });
          }
        }
      }

      duringRenderCall = true;
      let result;
      if (render) result = render.call(this);
      duringRenderCall = false;
      return result;
    }; // Schedule rerender when observable objects change


    this.render = (0, _observerUtil.observe)(this.render, {
      scheduler: () => this.setState({}),
      lazy: true
    });
    let componentDidMount; // Storage for this.componentDidMount

    let rootNode; // Stores used prop keys in a tree for fast traversal. A node is a map of child key to child node.

    if (this.mounted || this.componentDidMount) {
      componentDidMount = this.componentDidMount;

      this.componentDidMount = () => {
        let parentPropToParentNode = new Map(); // References to parent nodes containing used prop keys

        let propToProxy = new Map();
        let handlers = {
          get(target, key, receiver) {
            let result = Reflect.get(target, key, receiver); // See https://stackoverflow.com/a/39003741

            let parentNode = parentPropToParentNode.get(target);

            if (!parentNode.has(key)) {
              let node = new Map();
              parentNode.set(key, node);
              parentPropToParentNode.set(result, node);
            }

            if (typeof result == 'object' && result != null) {
              // If proxy is already created, return it instead of making a new proxy
              let proxy = propToProxy.get(result);
              if (proxy) return proxy;
              proxy = new Proxy(result, handlers);
              propToProxy.set(result, proxy);
              return proxy;
            }

            return result;
          }

        };
        let storedProps = this.props;
        let mutableProps = Object.assign({}, this.props); // Only shallow copy needed since freeze and seal are shallow

        rootNode = new Map();
        parentPropToParentNode.set(mutableProps, rootNode);
        this.props = new Proxy(mutableProps, handlers);
        let result;
        if (this.mounted) result = this.mounted();else if (componentDidMount) result = componentDidMount();
        this.props = storedProps;
        return result;
      };
    } // Disable rerender when component is about to be unmounted


    const componentWillUnmount = this.componentWillUnmount;

    this.componentWillUnmount = () => {
      let instances = Component._instances.get(this.constructor.name);

      instances.splice(instances.indexOf(this), 1);
      (0, _observerUtil.unobserve)(this.render);
      if (this.beforeUnmount) return this.beforeUnmount();
      if (componentWillUnmount) return componentWillUnmount();
    }; // Calls componentWillUnmount and componentDidMount when props that are used in componentDidMount changes


    const componentDidUpdate = this.componentDidUpdate;

    this.componentDidUpdate = (prevProps, prevState, snapshot) => {
      if (!this.updated && // This method might already handle calling componentWillUnmount or componentDidMount
      !componentDidUpdate && // This method might already handle calling componentWillUnmount or componentDidMount
      this.componentDidMount && this.props != prevProps) {
        const equal = (prevProps, props, parentNode) => {
          for (const [key, node] of parentNode) {
            const prevValue = prevProps[key];
            const value = props[key];
            if (typeof prevValue == 'object' && typeof value == 'object' ? prevValue != null && value == null || prevValue == null && value != null : prevValue !== value) return false;
            if (!equal(prevProps[key], props[key], node)) return false;
          }

          return true;
        };

        if (!equal(prevProps, this.props, rootNode)) {
          if (this.beforeUnmount) this.beforeUnmount();else if (componentWillUnmount) componentWillUnmount();
          this.componentDidMount();
        }
      }

      if (this.updated) return this.updated(prevProps, prevState, snapshot);
      if (componentDidUpdate) return componentDidUpdate.call(this, prevProps, prevState, snapshot);
    };
  }

}

exports.Component = Component;